# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""SKOS import overrides"""

import sys

from cubicweb.dataimport.stores import NoHookRQLObjectStore

from cubicweb_skos import sobjects as skos

from . import SAEMMetadataGenerator


def _store(cnx, **kwargs):
    metagen = SAEMMetadataGenerator(cnx, **kwargs)
    # XXX detect if import is started through source-sync, because the massive store isn't usable
    # when other connections are active on the database
    if 'source-sync' in sys.argv and cnx.repo.system_source.dbdriver == 'postgres':
        from cubicweb.dataimport.massive_store import MassiveObjectStore
        return MassiveObjectStore(cnx, metagen=metagen, eids_seq_range=1000)
    else:
        return NoHookRQLObjectStore(cnx, metagen=metagen)


class RDFSKOSImportService(skos.RDFSKOSImportService):
    """service overriden from skos cube to provide functionnalities handled in saem_ref's hooks"""

    def _do_import(self, stream, import_log, **kwargs):
        return super(RDFSKOSImportService, self)._do_import(stream, import_log,
                                                            store=_store(self._cw), **kwargs)


class LCSVSKOSImportService(skos.LCSVSKOSImportService):
    """service overriden from skos cube to provide functionnalities handled in saem_ref's hooks"""

    def _do_import(self, stream, import_log, **kwargs):
        scheme = self._cw.find('ConceptScheme', cwuri=kwargs['scheme_uri']).one()
        naa_what = scheme.cw_adapt_to('IArkNAALocator').naa_what()
        store = _store(self._cw, naa_what=naa_what)
        return super(LCSVSKOSImportService, self)._do_import(stream, import_log,
                                                             store=store, **kwargs)


class SKOSParser(skos.SKOSParser):
    """parser overriden from skos cube to provide functionnalities handled in saem_ref's hooks"""

    def _do_import(self, url, raise_on_error):
        return super(SKOSParser, self)._do_import(url, raise_on_error=raise_on_error,
                                                  store=_store(self._cw, source=self.source))


def registration_callback(vreg):
    vreg.register_and_replace(RDFSKOSImportService, skos.RDFSKOSImportService)
    vreg.register_and_replace(LCSVSKOSImportService, skos.LCSVSKOSImportService)
    vreg.register_and_replace(SKOSParser, skos.SKOSParser)



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Adapters to export content to RDF"""

from cubicweb.predicates import is_instance

from cubicweb_skos import rdfio, entities as skos

from .. import permanent_url


def _prov_mapping(reg, etype):
    """Register RDF mapping for PROV-O entity types related to some entity type (prov:Agent).
    """
    reg.register_prefix('prov', 'http://www.w3.org/ns/prov#')
    reg.register_etype_equivalence('Activity', 'prov:Activity')
    reg.register_attribute_equivalence('Activity', 'type', 'prov:type')
    reg.register_attribute_equivalence('Activity', 'description', 'prov:label')
    reg.register_attribute_equivalence('Activity', 'start', 'prov:startedAtTime')
    reg.register_attribute_equivalence('Activity', 'end', 'prov:endedAtTime')
    reg.register_relation_equivalence('Activity', 'associated_with', 'CWUser',
                                      'prov:wasAssociatedWith')
    reg.register_relation_equivalence('Activity', 'generated', etype, 'prov:generated')
    reg.register_relation_equivalence('Activity', 'used', etype, 'prov:used')


def _base_rdf_mapping(reg, etype, kind):
    reg.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
    reg.register_prefix('org', 'http://www.w3.org/ns/org#')
    reg.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
    reg.register_etype_equivalence(etype, kind)
    reg.register_attribute_equivalence(
        etype, lambda entity: u'ark:/' + entity.ark, 'dc:identifier')


def _complete_agent_rdf_mapping(reg, etype):
    reg.register_prefix('dct', 'http://purl.org/dc/terms/')
    reg.register_prefix('schema_org', 'http://schema.org/')
    reg.register_prefix('vcard', 'http://www.w3.org/2006/vcard/ns#')
    reg.register_attribute_equivalence(etype, 'creation_date', 'dct:created')
    reg.register_attribute_equivalence(etype, 'modification_date', 'dct:modified')


def _complete_authority_record_rdf_mapping(reg):
    _complete_agent_rdf_mapping(reg, 'AuthorityRecord')
    _prov_mapping(reg, 'AuthorityRecord')
    reg.register_attribute_equivalence('AuthorityRecord', lambda x: x.dc_title(), 'foaf:name')
    reg.register_attribute_equivalence('AuthorityRecord', 'start_date', 'schema_org:startDate')
    reg.register_attribute_equivalence('AuthorityRecord', 'end_date', 'schema_org:endDate')
    reg.register_etype_equivalence('PostalAddress', 'vcard:Location')
    reg.register_attribute_equivalence('PostalAddress', 'street', 'vcard:street-address')
    reg.register_attribute_equivalence('PostalAddress', 'postalcode', 'vcard:postal-code')
    reg.register_attribute_equivalence('PostalAddress', 'city', 'vcard:locality')
    reg.register_attribute_equivalence('PostalAddress', 'country', 'vcard:country-name')
    reg.register_attribute_equivalence('PostalAddress', 'state', 'vcard:region')


class AgentRDFListAdapter(skos.AbstractRDFAdapter):
    __regid__ = 'RDFList'
    __select__ = is_instance('Agent')

    def register_rdf_mapping(self, reg):
        _base_rdf_mapping(reg, self.entity.cw_etype, 'foaf:Person')
        reg.register_attribute_equivalence(
            self.entity.cw_etype, 'name', 'foaf:name')


class OrgRDFListAdapter(skos.AbstractRDFAdapter):
    __regid__ = 'RDFList'
    __select__ = is_instance('Organization', 'OrganizationUnit')

    def register_rdf_mapping(self, reg):
        etype = self.entity.cw_etype
        _base_rdf_mapping(reg, etype, 'org:{0}'.format(etype))
        reg.register_attribute_equivalence(etype, 'name', 'dc:title')


class AgentRDFPrimaryAdapter(AgentRDFListAdapter):
    __regid__ = 'RDFPrimary'
    __select__ = is_instance('Agent')

    def register_rdf_mapping(self, reg):
        super(AgentRDFPrimaryAdapter, self).register_rdf_mapping(reg)
        _complete_agent_rdf_mapping(reg, self.entity.cw_etype)


class OrganizationUnitRDFPrimaryAdapter(OrgRDFListAdapter):
    __regid__ = 'RDFPrimary'
    __select__ = is_instance('OrganizationUnit')

    def register_rdf_mapping(self, reg):
        super(OrganizationUnitRDFPrimaryAdapter, self).register_rdf_mapping(reg)
        _complete_agent_rdf_mapping(reg, self.entity.cw_etype)
        reg.register_relation_equivalence(
            'OrganizationUnit', 'contact_point', 'Agent',
            'schema_org:contactPoint')
        reg.register_relation_equivalence(
            'OrganizationUnit', 'authority', 'Organization',
            'org:unitOf')

    def fill(self, graph):
        super(OrganizationUnitRDFPrimaryAdapter, self).fill(graph)
        reg = self.registry
        # Export archival roles for the agent
        for archival_role in self.entity.archival_role:
            graph.add(graph.uri(permanent_url(self.entity)),
                      graph.uri(reg.normalize_uri('vcard:role')),
                      archival_role.name)
        # Export contact agent
        if self.entity.contact_point:
            contact_agent = self.entity.contact_point[0]
            contact_agent.cw_adapt_to('RDFPrimary').fill(graph)
        # Export related organization
        if self.entity.authority:
            org = self.entity.authority[0]
            org.cw_adapt_to('RDFPrimary').fill(graph)


class OrganizationRDFPrimaryAdapter(OrgRDFListAdapter):
    __regid__ = 'RDFPrimary'
    __select__ = is_instance('Organization')

    def register_rdf_mapping(self, reg):
        super(OrganizationRDFPrimaryAdapter, self).register_rdf_mapping(reg)
        _complete_agent_rdf_mapping(reg, self.entity.cw_etype)
        reg.register_prefix('saem', 'http://www.logilab.org/saem/0#')
        reg.register_relation_equivalence(
            'Organization', 'archival_unit', 'OrganizationUnit',
            'saem:archivalUnit')
        reg.register_relation_equivalence(
            'Organization', 'archival_authority', 'Organization',
            'saem:archivalAuthority')


class AuthorityRecordRDFPrimaryAdapter(skos.AbstractRDFAdapter):
    __regid__ = 'RDFPrimary'
    __select__ = is_instance('AuthorityRecord')

    hierarchical_role = 'http://www.logilab.org/saem/hierarchical_role'
    association_role = 'http://www.logilab.org/saem/association_role'

    def register_rdf_mapping(self, reg):
        """RDF mapping for AuthorityRecord entity type."""
        reg.register_prefix('org', 'http://www.w3.org/ns/org#')
        reg.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
        kind = {'person': 'foaf:Person',
                'authority': 'org:OrganizationUnit',
                'family': 'foaf:Group'}.get(self.entity.kind, 'foaf:Person')
        _base_rdf_mapping(reg, 'AuthorityRecord', kind)
        _complete_authority_record_rdf_mapping(reg)

    def accept(self, entity):
        """Return True if the entity should be recursivly added to the graph."""
        return entity.cw_etype == 'Activity'

    def fill(self, graph):
        super(AuthorityRecordRDFPrimaryAdapter, self).fill(graph)
        reg = self.registry
        generator = rdfio.RDFGraphGenerator(graph)
        # Export addresses
        for place in self.entity.reverse_place_agent:
            for address in place.place_address:
                graph.add(graph.uri(permanent_url(self.entity)),
                          graph.uri(reg.normalize_uri('vcard:hasAddress')),
                          graph.uri(permanent_url(address)))
                if place.role:
                    graph.add(graph.uri(permanent_url(address)),
                              graph.uri(reg.normalize_uri('vcard:role')),
                              place.role)
                generator.add_entity(address, reg)
        # rtype name: (subj property, obj property)
        rtype_to_properties = {
            'reverse_chronological_successor': ('dct:replaces', 'dct:isReplacedBy'),
            'reverse_chronological_predecessor': ('dct:isReplacedBy', 'dct:replaces'),
            'reverse_hierarchical_parent': ('org:organization', 'org:member'),
            'reverse_hierarchical_child': ('org:member', 'org:organization'),
        }
        # Export chronological relations
        relations_generator = _iter_on_relations(
            self.entity,
            [('reverse_chronological_successor', 'chronological_predecessor'),
             ('reverse_chronological_predecessor', 'chronological_successor')])
        for rtype_name, relation, related_record in relations_generator:
            to_predicate, from_predicate = rtype_to_properties[rtype_name]
            graph.add(graph.uri(permanent_url(related_record)),
                      graph.uri(reg.normalize_uri(from_predicate)),
                      graph.uri(permanent_url(self.entity)))
            graph.add(graph.uri(permanent_url(self.entity)),
                      graph.uri(reg.normalize_uri(to_predicate)),
                      graph.uri(permanent_url(related_record)))
        # Export hierarchical relations
        relations_generator = _iter_on_relations(
            self.entity,
            [('reverse_hierarchical_parent', 'hierarchical_child'),
             ('reverse_hierarchical_child', 'hierarchical_parent')])
        self._add_membership_to_graph(graph, self.hierarchical_role, relations_generator,
                                      rtype_to_properties)
        # Export association relations
        relations_generator = _iter_on_relations(
            self.entity,
            [('reverse_association_from', 'association_to'),
             ('reverse_association_to', 'association_from')])
        # As this is an association relation, it makes no sense to have an agent being the
        # organization rather than the other one. That is why we "duplicate" Membership
        # properties: each agent will be both known as a member and an organization.
        relations = list(relations_generator)
        self._add_membership_to_graph(
            graph, self.association_role, relations,
            {'reverse_association_from': ('org:organization', 'org:member'),
             'reverse_association_to': ('org:member', 'org:organization')})
        self._add_membership_to_graph(
            graph, self.association_role, relations,
            {'reverse_association_from': ('org:member', 'org:organization'),
             'reverse_association_to': ('org:organization', 'org:member')})

    def _add_membership_to_graph(self, graph, role, relations, rtype_to_properties):
        reg = self.registry
        for rtype_name, relation, related_record in relations:
            relation_url = permanent_url(relation)
            graph.add(graph.uri(relation_url),
                      graph.uri(reg.normalize_uri(rtype_to_properties[rtype_name][0])),
                      graph.uri(permanent_url(self.entity)))
            graph.add(graph.uri(relation_url),
                      graph.uri(reg.normalize_uri(rtype_to_properties[rtype_name][1])),
                      graph.uri(permanent_url(related_record)))
            graph.add(graph.uri(relation_url),
                      graph.uri(reg.normalize_uri('org:role')),
                      graph.uri(role))
            if relation.start_date or relation.end_date:
                interval_resource = reg.normalize_uri(relation_url + '#timeInterval')
                graph.add(graph.uri(relation_url),
                          graph.uri(reg.normalize_uri('org:memberDuring')),
                          graph.uri(interval_resource))
            if relation.start_date:
                graph.add(graph.uri(interval_resource),
                          graph.uri(reg.normalize_uri('schema_org:startDate')),
                          relation.start_date)
            if relation.end_date:
                graph.add(graph.uri(interval_resource),
                          graph.uri(reg.normalize_uri('schema_org:endDate')),
                          relation.end_date)


def _iter_on_relations(entity, relation_descriptions):
    """yield (rtype_name, relation, related_entity)

    `relation_descriptions` is a list of (rtype_name, reverse_rtype_name)
    """
    for rtype_name, reverse_rtype_name in relation_descriptions:
        for relation in getattr(entity, rtype_name):
            for related_entity in getattr(relation, reverse_rtype_name):
                yield rtype_name, relation, related_entity


_ACTIVITY_ATTRIBUTES = 'cwuri description type start end'.split()


class ConceptSchemeRDFPrimaryAdapter(skos.ConceptSchemeRDFPrimaryAdapter):

    # prefetch 'ark' attribute of concept for RDF export
    skos.ConceptSchemeRDFPrimaryAdapter.concept_attributes.append('ark')

    def register_rdf_mapping(self, reg):
        super(ConceptSchemeRDFPrimaryAdapter, self).register_rdf_mapping(reg)
        _prov_mapping(reg, 'ConceptScheme')

    def accept(self, entity):
        """Return True if the entity should be recursivly added to the graph."""
        return entity.cw_etype == 'Activity'

    def warm_caches(self):
        super(ConceptSchemeRDFPrimaryAdapter, self).warm_caches()
        scheme = self.entity
        cnx = self._cw

        activities, rset = _warm_activity_cache(cnx, scheme, 'CS identity X, A generated X')
        scheme._cw_related_cache['generated_object'] = (rset, activities)
        scheme._cw_related_cache['used_object'] = (rset, activities)

        activities, rset = _warm_activity_cache(cnx, scheme, 'X in_scheme CS, A generated X')
        concepts = tuple(rset.entities(1))
        skos.cache_entities_relations(concepts, rset, 'generated', 'object',
                                      entity_col=0, target_col=1)
        skos.cache_entities_relations(concepts, rset, 'used', 'object',
                                      entity_col=0, target_col=1)


def _warm_activity_cache(cnx, scheme, rql_expr):
    activities_rql = skos._select_attributes(
        'Any X,A WHERE CS eid %(cs)s,' + rql_expr,
        'A', _ACTIVITY_ATTRIBUTES)
    rset = cnx.execute(activities_rql, {'cs': scheme.eid})
    activities = tuple(rset.entities(1))
    skos.cache_entities_relations(activities, rset, 'generated', 'subject')
    skos.cache_entities_relations(activities, rset, 'used', 'subject')
    user_rset = cnx.execute('Any U,A WHERE A associated_with U, CS eid %(cs)s, ' + rql_expr,
                            {'cs': scheme.eid})
    skos.cache_entities_relations(activities, user_rset, 'associated_with', 'subject')
    return activities, rset


class ConceptRDFPrimaryAdapter(skos.ConceptRDFPrimaryAdapter):

    def register_rdf_mapping(self, reg):
        super(ConceptRDFPrimaryAdapter, self).register_rdf_mapping(reg)
        _prov_mapping(reg, 'Concept')

    def accept(self, entity):
        """Return True if the entity should be recursivly added to the graph."""
        return entity.cw_etype == 'Activity'


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, [ConceptSchemeRDFPrimaryAdapter,
                                                     ConceptRDFPrimaryAdapter])
    vreg.register_and_replace(ConceptSchemeRDFPrimaryAdapter, skos.ConceptSchemeRDFPrimaryAdapter)
    vreg.register_and_replace(ConceptRDFPrimaryAdapter, skos.ConceptRDFPrimaryAdapter)



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as publishged by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref tests for SEDA cube integration"""

import re
import unittest

from six import text_type

from cubicweb.devtools import PostgresApptestConfiguration
from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_seda.entities.profile_generation import SEDA1XSDExport, SEDA2RelaxNGExport

from cubicweb_saem_ref.entities import seda  # noqa - trigger monkey-patch

import testutils


def setUpModule():
    testutils.startpgcluster(__file__)


class CWURIURLTC(unittest.TestCase):

    def test(self):

        class FakeEntity(object):
            def __init__(self, cwuri):
                self.cwuri = cwuri
                self._cw = self

            def build_url(self, uri):
                return 'http://thistest/' + uri

        self.assertEqual(SEDA1XSDExport.cwuri_url(FakeEntity('ark:12/3')),
                         'http://thistest/ark:12/3')
        self.assertEqual(SEDA1XSDExport.cwuri_url(FakeEntity('http://othertest/ark:12/3')),
                         'http://othertest/ark:12/3')
        self.assertEqual(SEDA1XSDExport.cwuri_url(FakeEntity('http://othertest/whatever')),
                         'http://othertest/whatever')
        self.assertEqual(SEDA1XSDExport.cwuri_url(FakeEntity('whatever')),
                         'whatever')

        self.assertEqual(SEDA2RelaxNGExport.cwuri_url(FakeEntity('ark:12/3')),
                         'http://thistest/ark:12/3')
        self.assertEqual(SEDA2RelaxNGExport.cwuri_url(FakeEntity('http://othertest/ark:12/3')),
                         'http://othertest/ark:12/3')
        self.assertEqual(SEDA2RelaxNGExport.cwuri_url(FakeEntity('http://othertest/whatever')),
                         'http://othertest/whatever')
        self.assertEqual(SEDA2RelaxNGExport.cwuri_url(FakeEntity('whatever')),
                         'whatever')


class SEDAExportTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def iter_seda_xsd(self, transfer):
        """Yield XSD representations of the transfer for SEDA 0.2 then 1.0,
        within a subTest.
        """
        for adapter_id in ('SEDA-0.2.xsd', 'SEDA-1.0.xsd'):
            with self.subTest(adapter=adapter_id):
                adapter = transfer.cw_adapt_to(adapter_id)
                yield adapter.dump()

    def test_include_profile_ark(self):
        with self.admin_access.client_cnx() as cnx:
            transfer = testutils.seda_transfer(cnx)
            testutils.create_archive_unit(transfer)
            cnx.commit()
            for xml in self.iter_seda_xsd(transfer):
                self.assertIn('ArchivalProfile', xml)
                self.assertIn('ark:/' + transfer.ark, xml)

    def test_include_agency_ark(self):
        with self.admin_access.client_cnx() as cnx:
            transfer = testutils.seda_transfer(cnx)
            _, _, unit_alt_seq = testutils.create_archive_unit(transfer)
            agency = cnx.create_entity('SEDAOriginatingAgency',
                                       seda_originating_agency_from=unit_alt_seq)

            # ensure it doesn't crash if agency is not tight to an authority record
            for xml in self.iter_seda_xsd(transfer):
                self.assertIn('OriginatingAgency', xml)

            # ensure we get the ark as agency identifier
            record = testutils.authority_record(cnx, u'DGSI')
            agency.cw_set(seda_originating_agency_to=record)
            for xml in self.iter_seda_xsd(transfer):
                self.assertIn('ark:/' + record.ark, xml)

    def test_TransferringAgencyArchiveIdentifier_present_rng02(self):
        with self.admin_access.cnx() as cnx:
            transfer = testutils.seda_transfer(cnx)
            testutils.create_archive_unit(transfer)
            cnx.commit()
            adapter = transfer.cw_adapt_to('SEDA-0.2.rng')
            root = adapter.dump_etree()
            # We expect a rng:element with
            # name=TransferringAgencyArchiveIdentifier,
            results = root.xpath('//*[@name="TransferringAgencyArchiveIdentifier"]')
            self.assertEqual(len(results), 1)
            tag = results[0]
            # there should be no rng:value tag.
            self.assertIsNone(tag.find('rng:value', namespaces=root.nsmap))

    def test_TransferringAgencyArchiveIdentifier_present_xsd(self):
        with self.admin_access.cnx() as cnx:
            transfer = testutils.seda_transfer(cnx)
            testutils.create_archive_unit(transfer)
            cnx.commit()
            adapter = transfer.cw_adapt_to('SEDA-0.2.xsd')
            root = adapter.dump_etree()
            # We expect a rng:element with
            # name=TransferringAgencyArchiveIdentifier,
            results = root.xpath('//*[@name="TransferringAgencyArchiveIdentifier"]')
            self.assertEqual(len(results), 1)


class CloneImportTC(CubicWebTC):
    """Tests for cubicweb-seda's seda.doimport controller, making sure it
    works in SAEM-Ref as well.

    Copied and adapted eponymous class in cubicweb-seda/test/test_views.py.
    """

    configcls = PostgresApptestConfiguration

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            transfer = testutils.setup_profile(cnx, title=u'test')
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(None, cnx=cnx,
                                                                         user_cardinality=u'1',
                                                                         user_annotation=u'plop')
            self.transfer_eid = transfer.eid
            self.unit_eid = unit.eid
            self.unit_alt_seq_eid = unit_alt_seq.eid
            cnx.commit()

    def test_import_one_entity(self):
        params = dict(eid=text_type(self.transfer_eid),
                      cloned=text_type(self.unit_eid))
        with self.admin_access.web_request(**params) as req:
            path, _ = self.expect_redirect_handle_request(
                req, 'seda.doimport')
            match = re.match(r'^ark:/(0/rf.{7}g)$', path)
            assert match, path
            ark = match.group(1)
            rset = req.execute('Any X WHERE X seda_archive_unit P, P ark %(ark)s',
                               {'ark': ark})
            self.assertTrue(rset)
            imported_unit = rset.one()
            self.assertNotEqual(req.entity_from_eid(self.unit_eid).ark,
                                imported_unit.ark)
            # Make sure archive unit's ARK is qualified w.r.t. its parent
            # archive transfer's ARK.
            self.assertTrue(imported_unit.ark.startswith(ark), imported_unit)

    def test_create_nested_archive_unit(self):
        with self.admin_access.cnx() as cnx:
            unit_alt_seq = cnx.entity_from_eid(self.unit_alt_seq_eid)
            unit, _, _ = testutils.create_archive_unit(
                unit_alt_seq, cnx=cnx,
                ark_naa=None,
                user_cardinality=u'1',
                user_annotation=u'nested')
            cnx.commit()
            self.assertTrue(unit.ark)
            self.assertFalse(unit.ark_naa)
            parent = cnx.entity_from_eid(self.unit_eid)
            # Nested archive unit has a qualified ARK from its parent archive
            # unit.
            self.assertTrue(unit.ark.startswith(parent.ark), unit.ark)

    def test_clone_with_nested_archive_units(self):
        with self.admin_access.cnx() as cnx:
            transfer = cnx.entity_from_eid(self.transfer_eid)
            _, _, unit_alt_seq = testutils.create_archive_unit(
                transfer, cnx=cnx,
                ark_naa=None,
                user_cardinality=u'1',
                user_annotation=u'top-level')
            cnx.commit()
            assert transfer.reverse_seda_archive_unit
            testutils.create_archive_unit(
                unit_alt_seq, cnx=cnx,
                ark_naa=None,
                user_cardinality=u'1',
                user_annotation=u'nested')
            cnx.commit()
            rset = cnx.execute('Any X,Y WHERE X seda_archive_unit Y')
            self.assertEqual(len(rset), 2, rset)
            transfer.cw_adapt_to('IWorkflowable').fire_transition(u'publish')
            cnx.commit()
            testutils.setup_profile(cnx, title=u'Clone',
                                    new_version_of=transfer)
            cnx.commit()
            rset = cnx.execute('Any X,Y WHERE X seda_archive_unit Y')
            self.assertEqual(len(rset), 4, rset)


if __name__ == '__main__':
    unittest.main()



"""ReST extensions"""

from pkg_resources import Requirement, resource_filename, resource_exists
import tempfile
import urllib2

from docutils.parsers.rst.directives import misc, register_directive


class PkgResourcesInclude(misc.Include):
    """Implement the `pkg-resources-include` directive, similar to the `include`
    directive but taking a pypi project name as first argument and a data file
    within the package as second argument.

    Actual file location will be computed using pkg_resources facilities.

    Example:

        .. pkg-resources-include:: cubicweb-seda doc/profils.rst

    """

    required_arguments = 2

    def run(self):
        project, fpath = self.arguments

        if not resource_exists(Requirement.parse(project), fpath):
            raise Exception('Resource %s not found in %s' % (fpath, project))

        filename = resource_filename(Requirement.parse(project), fpath)
        self.arguments = [filename]
        return super(PkgResourcesInclude, self).run()


register_directive('pkg-resources-include', PkgResourcesInclude)


class IncludeURL(misc.Include):
    """Implement the `include-url` directive, similar to the `include`
    directive but taking an URL as argument.

    Example:

        .. include-url:: https://framagit.org/saemproject/saem-client/raw/master/README.rst
    """

    def run(self):
        url, = self.arguments

        input_stream = urllib2.urlopen(url)
        with tempfile.NamedTemporaryFile() as output_stream:
            output_stream.write(input_stream.read())
            output_stream.flush()
            self.arguments = [output_stream.name]
            return super(IncludeURL, self).run()


register_directive('include-url', IncludeURL)



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref SKOS entity's classes"""

from logilab.common.decorators import cachedproperty

from cubicweb.predicates import is_instance, score_entity
from cubicweb.entities import fetch_config

from cubicweb_skos import entities as skos
from cubes.oaipmh.entities import ETypeOAISetSpec, RelatedEntityOAISetSpec

from . import oai


class ConceptScheme(skos.ConceptScheme):
    fetch_attrs, cw_fetch_order = fetch_config(('title', 'cwuri', 'ark',
                                                'description', 'description_format'))

    @cachedproperty
    def top_concepts_rset(self):
        rset = self._cw.execute(
            'Any C,CU,CA,CD,CE WHERE C cwuri CU, C ark CA, C definition CD, C example CE, '
            'C in_scheme X, NOT C broader_concept SC, X eid %(x)s',
            {'x': self.eid})
        if len(rset) > self._cw.property_value('navigation.page-size'):
            return rset  # optimizations below are not worth it
        # prefill 'preferred_label' relation cache for the above concepts
        labels_rset = self._cw.execute(
            'Any L,LL,LK,LLC,C ORDERBY C WHERE '
            'L label LL, L kind LK, L language_code LLC, C preferred_label L, '
            'C in_scheme X, NOT C broader_concept SC, X eid %(x)s',
            {'x': self.eid})
        entities = tuple(rset.entities())
        skos.cache_entities_relations(entities, labels_rset, 'preferred_label', 'subject',
                                      entity_col=4, target_col=0)
        narrower_rset = self._cw.execute(
            'Any C,CU,CA,CD,CE,BC ORDERBY BC WHERE '
            'C cwuri CU, C ark CA, C definition CD, C example CE, '
            'BC in_scheme X, C broader_concept BC, NOT BC broader_concept SC, X eid %(x)s',
            {'x': self.eid})
        skos.cache_entities_relations(entities, narrower_rset, 'broader_concept', 'object',
                                      entity_col=5, target_col=0)
        # return the original rset
        return rset


class Concept(skos.Concept):
    fetch_attrs, cw_fetch_order = fetch_config(('cwuri', 'ark', 'definition', 'example'))


class ConceptSchemeOAIPMHRecordAdapter(oai.OAIPMHActiveRecordAdapter):
    """OAI-PMH adapter for ConceptScheme entity type."""
    __select__ = oai.OAIPMHActiveRecordAdapter.__select__ & is_instance('ConceptScheme')
    metadata_formats = {'rdf': (oai.RDF_METADATAFORMAT, 'list.rdf')}

    @classmethod
    def set_definition(cls):
        return oai.PublicETypeOAISetSpec(
            'ConceptScheme', cls.identifier_attribute)


class ConceptInPublicSchemeETypeOAISetSpec(ETypeOAISetSpec):
    """OAI-PMH set specifier matching on "Concept" entity type and fetching
    concepts related to scheme not in "draft" state.
    """

    def __init__(self):
        super(ConceptInPublicSchemeETypeOAISetSpec, self).__init__(
            'Concept', 'ark')

    def setspec_restrictions(self, value=None):
        return 'X in_scheme Y, Y in_state ST, NOT ST name "draft"', {}


class ConceptInPublicSchemeRelatedEntityOAISetSpec(RelatedEntityOAISetSpec):
    """OAI-PMH second-level set specifier to match "Concept" entity type
    related to scheme not in "draft" state.
    """

    def __init__(self):
        super(ConceptInPublicSchemeRelatedEntityOAISetSpec, self).__init__(
            'in_scheme', 'ConceptScheme', 'ark',
            description=u'A concept in scheme identified by {0}')

    def setspecs(self, cnx):
        rset = cnx.execute('Any A WHERE X is ConceptScheme, X ark A,'
                           '            X in_state ST, NOT ST name "draft"')
        for value, in rset.rows:
            yield value, self.description.format(value)


def not_in_draft_scheme(concept):
    """Return True if `concept` is in a ConceptScheme not in "draft" state.
    """
    return concept.in_scheme[0].in_state[0].name != 'draft'


class ConceptOAIPMHRecordAdapter(oai.ArkOAIPMHRecordAdapter):
    """OAI-PMH adapter for ConceptScheme entity type."""
    __select__ = (oai.ArkOAIPMHRecordAdapter.__select__ & is_instance('Concept')
                  & score_entity(not_in_draft_scheme))
    metadata_formats = {'rdf': (oai.RDF_METADATAFORMAT, 'primary.rdf')}

    @classmethod
    def set_definition(cls):
        specifier = ConceptInPublicSchemeETypeOAISetSpec()
        specifier['in_scheme'] = ConceptInPublicSchemeRelatedEntityOAISetSpec()
        return specifier



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref views related to SEDA"""

from cubicweb.predicates import (
    has_related_entities,
    is_instance,
    objectify_predicate,
)
from cubicweb.web import formwidgets as fw
from cubicweb.web.views import uicfg

from cubicweb_seda.views import (
    CONTENT_ETYPE,
    archivetransfer,
    archiveunit,
    rtags_from_rtype_role_targets,
    simplified
)


affk = uicfg.autoform_field_kwargs
afs = uicfg.autoform_section
pvs = uicfg.primaryview_section

# also hide transferring and archival agency
for rtype in ('seda_transferring_agency', 'seda_archival_agency'):
    afs.tag_subject_of(('SEDAArchiveTransfer', rtype, '*'), 'main', 'hidden')
    pvs.tag_subject_of(('SEDAArchiveTransfer', rtype, '*'), 'hidden')

pvs.tag_object_of(('*', 'use_profile', '*'), 'hidden')
afs.tag_object_of(('*', 'use_profile', '*'), 'main', 'hidden')

pvs.tag_attribute(('SEDABinaryDataObject', 'filename'), 'hidden')
afs.tag_attribute(('SEDABinaryDataObject', 'filename'), 'main', 'hidden')

# Have "seda_archive_unit" inserted in attributes section so that it appears
# in cw_edited and ARK generation hook works.
afs.tag_subject_of(('SEDAArchiveUnit', 'seda_archive_unit', '*'), 'main', 'attributes')
# Though we don't want to see the field in form, so hide the widget.
affk.set_field_kwargs('SEDAArchiveUnit', 'seda_archive_unit', widget=fw.HiddenInput())


archivetransfer.ArchiveTransferTabbedPrimaryView.tabs.append('saem.lifecycle_tab')


@objectify_predicate
def creating_seda_archive_unit(cls, req, **kwargs):
    for linkto in req.list_form_param('__linkto'):
        rtype, _, role = linkto.split(':')
        if (rtype, role) == ('seda_archive_unit', 'subject'):
            return 1
    return 0


nested_archiveunit_afs = afs.derive(
    __name__,
    is_instance('SEDAArchiveUnit')
    & (has_related_entities('seda_archive_unit')
       | creating_seda_archive_unit())
)
nested_archiveunit_afs.tag_subject_of(('SEDAArchiveUnit', 'ark_naa', '*'),
                                      'main', 'hidden')

nested_archiveunit_pvs = pvs.derive(
    __name__,
    is_instance('SEDAArchiveUnit')
    & has_related_entities('seda_archive_unit')
)
nested_archiveunit_pvs.tag_subject_of(('SEDAArchiveUnit', 'ark_naa', '*'),
                                      'hidden')


# Hide "identifiant pour le service versant" which should be automatically
# filled.
class SimplifiedContentMainView(archiveunit.SimplifiedContentMainView):

    rtype_role_targets = archiveunit.SimplifiedContentMainView.rtype_role_targets[:]
    rtype_role_targets.remove(
        ('seda_transferring_agency_archive_unit_identifier', 'object', None)
    )
    rsection, display_ctrl = rtags_from_rtype_role_targets(CONTENT_ETYPE, rtype_role_targets)


simplified.simplified_afs.tag_object_of(
    ('*', 'seda_transferring_agency_archive_unit_identifier', CONTENT_ETYPE),
    'main', 'hidden',
)


class ContentMainView(archiveunit.ContentMainView):

    rtype_role_targets = archiveunit.ContentMainView.rtype_role_targets[:]
    rtype_role_targets.remove(
        ('seda_transferring_agency_archive_unit_identifier', 'object', None)
    )
    rsection, display_ctrl = rtags_from_rtype_role_targets(CONTENT_ETYPE, rtype_role_targets)


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, [
        ContentMainView,
        SimplifiedContentMainView,
    ])
    vreg.register_and_replace(ContentMainView, archiveunit.ContentMainView)
    vreg.register_and_replace(SimplifiedContentMainView, archiveunit.SimplifiedContentMainView)



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

from cubicweb.devtools import PostgresApptestConfiguration
from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_saem_ref.ark import (
    ARK_QUALIFIER_LENGTH,
)
import testutils


def setUpModule():
    testutils.startpgcluster(__file__)


class ArkGeneratorTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def test(self):
        what = '31400'
        with self.admin_access.repo_cnx() as cnx:
            generator = self.vreg['adapters'].select(
                'IARKGenerator', cnx, naa_what=what)
            ark = generator.generate_ark()
        self.assertTrue(ark.startswith(what + '/'), ark)
        name = ark[len(what) + 1:]
        self.assertTrue(testutils.match_ark_name(name), name)

    def _check_ark_matches_parent(self, ark, parent_ark):
        self.assertTrue(ark.startswith(parent_ark + '/'), ark)
        qualifier = ark[len(parent_ark) + 1:]
        self.assertEqual(len(qualifier), ARK_QUALIFIER_LENGTH)

    def test_concept(self):
        with self.admin_access.cnx() as cnx:
            scheme = testutils.setup_scheme(cnx, u'test')
            cnx.commit()
            scheme_ark = scheme.ark
            what = str(scheme.ark_naa[0].what)
            generator = self.vreg['adapters'].select(
                'IARKGenerator', cnx, naa_what=what, in_scheme=scheme)
            ark = generator.generate_ark()
            self._check_ark_matches_parent(ark, scheme_ark)

    def test_organizationunit(self):
        with self.admin_access.cnx() as cnx:
            authority = testutils.authority_with_naa(cnx)
            cnx.commit()
            authority_ark = authority.ark
            what = str(authority.ark_naa[0].what)
            generator = self.vreg['adapters'].select(
                'IARKGenerator', cnx, naa_what=what, authority=authority)
            ark = generator.generate_ark()
            self._check_ark_matches_parent(ark, authority_ark)

    def test_sedaarchiveunit(self):
        with self.admin_access.cnx() as cnx:
            transfer = testutils.setup_profile(cnx, title=u'test')
            cnx.commit()
            transfer_ark = transfer.ark
            what = str(transfer.ark_naa[0].what)
            generator = self.vreg['adapters'].select(
                'IARKGenerator', cnx, naa_what=what, seda_archive_unit=transfer)
            ark = generator.generate_ark()
            self._check_ark_matches_parent(ark, transfer_ark)


if __name__ == '__main__':
    import unittest
    unittest.main()



# coding: utf-8
import datetime

print 'create Logilab agent'
agent = create_entity('Agent', name=u'Logilab',
                      start_date=datetime.date(2000, 1, 1))
create_entity('PhoneNumber', number=u'05 62 17 16 42',
              reverse_phone_number=agent)
office = create_entity('AgentPlace', name=u'Toulouse', role=u'QG',
                       place_agent=agent)
create_entity('PostalAddress',
              street=u'1 av. de l\'Europe',
              postalcode=u'31400',
              city=u'Toulouse',
              reverse_place_address=office)
rql('SET A archival_role R WHERE R name in ("control", "deposit")')
rql('SET A agent_kind K WHERE K name "authority"')
commit()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref entity's classes for OAI-PMH protocol"""

from isodate import datetime_isoformat

from cubicweb.predicates import is_in_state, relation_possible, score_entity

from cubes.oaipmh import MetadataFormat
from cubes.oaipmh.entities import OAIPMHRecordAdapter, ETypeOAISetSpec, OAIComponent


# MetadataFormat instance for "rdf" with '<unknown>' schema since there
# appears to be no generic XML schema for RDF.
RDF_METADATAFORMAT = MetadataFormat(schema='<unknown>', namespace='rdf')


class SAEMOAIComponent(OAIComponent):
    """Override "oai" component to adjust ark identifier based on stored value
    (the former including the ark:/ prefix while the later does not).
    """
    deleted_handling = 'persistent'

    def match_identifier(self, identifier):
        prefix = 'ark:/'
        if identifier.startswith(prefix):
            identifier = identifier[len(prefix):]
        return super(SAEMOAIComponent, self).match_identifier(identifier)


class ArkOAIPMHRecordAdapter(OAIPMHRecordAdapter):
    """OAI-PMH Record using the "ark" attribute of entity as identifier.
    """

    __abstract__ = True
    __select__ = (OAIPMHRecordAdapter.__select__
                  & score_entity(lambda x: hasattr(x, 'ark')))
    identifier_attribute = 'ark'

    @property
    def identifier(self):
        return 'ark:/' + self.entity.ark


def in_state(state):
    """Predicate checking for workflowability and matching on `state`.
    """
    return relation_possible('in_state') & is_in_state(state)


def not_in_state(state):
    """Predicate checking for workflowability and not matching on `state`.
    """
    return relation_possible('in_state') & ~is_in_state(state)


class OAIPMHActiveRecordAdapter(ArkOAIPMHRecordAdapter):
    """OAI-PMH Record for "active" entities (not deprecated in our application).

    See http://www.openarchives.org/OAI/2.0/openarchivesprotocol.htm#DeletedRecords
    """

    __abstract__ = True
    __select__ = ArkOAIPMHRecordAdapter.__select__ & not_in_state('deprecated')


class OAIPMHDeletedRecordAdapter(ArkOAIPMHRecordAdapter):
    """OAI-PMH Record for "deleted" entities (deprecated in our application).

    See http://www.openarchives.org/OAI/2.0/openarchivesprotocol.htm#DeletedRecords

    Handle the representation of entity with an ARK which is deleted as an
    OAI-PMH record.
    """

    __abstract__ = True
    __select__ = ArkOAIPMHRecordAdapter.__select__ & in_state('deprecated')

    @property
    def datestamp(self):
        """For delete a record, the date and time that it was deleted on.
        """
        trinfo = self.entity.cw_adapt_to('IWorkflowable').latest_trinfo()
        assert trinfo.transition.name == 'deprecate', \
            'unexpected latest transition {0}'.format(trinfo.transition)
        return datetime_isoformat(trinfo.creation_date)

    deleted = True

    def metadata(self):
        return ''


class PublicETypeOAISetSpec(ETypeOAISetSpec):
    """OAI-PMH set specifier matching an entity type and entities in
    "published" or "deprecated" state.
    """

    def setspec_restrictions(self, value=None):
        """Add workflow state restrictions."""
        base, args = super(PublicETypeOAISetSpec, self).setspec_restrictions(value)
        restrictions = ', '.join([base, 'X in_state ST, NOT ST name "draft"'])
        return restrictions, args

    def query_for_identifier(self, identifier):
        qs, args = super(PublicETypeOAISetSpec, self).query_for_identifier(
            identifier)
        qs += ', X in_state ST, NOT ST name "draft"'
        return qs, args


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, (SAEMOAIComponent,))
    vreg.register_and_replace(SAEMOAIComponent, OAIComponent)



# coding: utf-8
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref test for views."""

import json
import os
from datetime import date
from tempfile import NamedTemporaryFile
try:
    import unittest2 as unittest
except ImportError:
    import unittest

from six import text_type

from yams.schema import role_name

from cubicweb.devtools import PostgresApptestConfiguration
from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_skos.rdfio import default_graph

from cubicweb_saem_ref.views.widgets import process_incomplete_date
from cubicweb_saem_ref.views.authorityrecord import SortableListViewMixin

import testutils


def setUpModule():
    testutils.startpgcluster(__file__)


def eid_from_ark(ark):
    name = ark.rsplit('/', 1)[-1]
    num = ''.join(c for c in name if c.isdigit())
    return int(num)


class SetDateTC(unittest.TestCase):

    def test_set_date_month_beginning(self):
        self.assertEqual(date(1997, 6, 5), process_incomplete_date("5/6/1997"))
        self.assertEqual(date(1997, 6, 5), process_incomplete_date("5/6/97"))
        self.assertEqual(date(2012, 6, 5), process_incomplete_date("5/6/12"))
        self.assertEqual(date(1997, 6, 1), process_incomplete_date("6/1997"))
        self.assertEqual(date(1997, 1, 1), process_incomplete_date("1997"))
        # XXX: behavior has changed in dateutil with dayfirst=True
        # https://github.com/dateutil/dateutil/commit/2d42e046d55b9fbbc0a2f41ce83fb8ec5de2d28b
        self.assertIn(process_incomplete_date("1997/6/5"), [date(1997, 6, 5),  # dateutil<=2.5.1
                                                            date(1997, 5, 6),  # dateutil>=2.5.2
                                                            ])

    def test_set_date_month_end(self):
        self.assertEqual(date(1994, 8, 28), process_incomplete_date("28/08/1994", True))
        self.assertEqual(date(1994, 8, 31), process_incomplete_date("08/1994", True))
        self.assertEqual(date(1994, 2, 28), process_incomplete_date("1994/02", True))
        self.assertEqual(date(1994, 12, 31), process_incomplete_date("1994", True))

    def test_set_date_failure(self):
        with self.assertRaises(ValueError) as cm:
            process_incomplete_date("31/02/2012")
        self.assertIn("day is out of range for month", str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            process_incomplete_date("20/14/2015")
        self.assertIn("month must be in 1..12", str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            process_incomplete_date("20")


class FuncViewsTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def test_citation_view(self):
        with self.admin_access.cnx() as cnx:
            bob = testutils.authority_record(cnx, u'bob')
            mandate = cnx.create_entity('Mandate', term=u'u',
                                        mandate_agent=bob)
            citation = cnx.create_entity('Citation',
                                         note=u'plop',
                                         reverse_has_citation=mandate)
            citation_uri = cnx.create_entity('Citation',
                                             note=u'plop',
                                             uri=u'http://pl.op/',
                                             reverse_has_citation=mandate)
            cnx.commit()
            without_link_html = citation.view('citation-link')
            with_link_html = citation_uri.view('citation-link')
            self.assertEqual(
                with_link_html,
                u'<a class="truncate" href="http://pl.op/" title="plop">plop</a>')
            self.assertEqual(without_link_html, u'<i class="truncate">plop</i>')

    def test_eac_import_user_has_no_naa(self):
        regid = 'eac.import'
        with self.admin_access.web_request() as req:
            req._ = lambda x: x + u' \xe9'
            output = self.view(regid, req=req, template=None)
            self.assertIn('be in an organization', output)
            self.assertIn('to access this functionnality', output)
        with self.admin_access.web_request() as req:
            req._ = lambda x: x + u' \xe9'
            authority = req.cnx.create_entity('Organization', name=u'dummy',
                                              ark=u'123/dummy')
            req.user.cw_set(authority=authority)
            req.cnx.commit()
            output = self.view(regid, req=req, template=None)
            self.assertIn('have an NAA configured', output)
            self.assertIn('to access this functionnality', output)

    def test_eac_import_ok(self):
        regid = 'eac.import'
        fname = 'FRAD033_EAC_00001_simplified.xml'
        with self.admin_access.web_request() as req:
            req.user.cw_set(authority=testutils.authority_with_naa(req.cnx))
            naa = req.cnx.create_entity('ArkNameAssigningAuthority', who=u'EAC', what=1)
            # simply test the form properly render and is well formed
            self.view(regid, req=req, template=None)
            fields = {'file': (fname, open(self.datapath('EAC/' + fname))),
                      'naa': text_type(naa.eid)}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            req.view(regid)
            # test AuthorityRecord have been created with a ARK from NAA
            # specified in form params
            arecord = req.find('AuthorityRecord', has_text=u'Gironde. Conseil général').one()
            self.assertTrue(arecord.ark.startswith('1/'), arecord.ark)

    def test_highlight_script_execution(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = testutils.setup_scheme(cnx, u'my thesaurus')
            cnx.commit()
            url = scheme.absolute_url(highlight='toto')
        self.assertIn(
            '$(document).ready(function(){$("h1, h2, h3, h4, h5, table tbody td")'
            '.highlight("toto");});}',
            self.http_publish(url)[0])

    def test_highlight_on_rql_plain_text_search_same_etype(self):
        with self.admin_access.client_cnx() as cnx:
            # we need two concepts to get a list view
            scheme = testutils.setup_scheme(cnx, u'my thesaurus', u'toto', u'tata toto')
            concept = scheme.reverse_in_scheme[0]
            cnx.commit()
            ark = concept.ark
        url = 'http://testing.fr/cubicweb/view?rql=toto&__fromsearchbox=1&subvid=tsearch'
        html = self.http_publish(url)[0]
        self.assertIn(
            '<a href="http://testing.fr/cubicweb/ark:/{}?highlight=toto" title="">'.format(ark),
            html)

    def test_skos_negociation(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'musique',
                                       ark_naa=testutils.naa(cnx))
            scheme.add_concept(u'pop')
            cnx.commit()
        with self.admin_access.web_request('conceptscheme',
                                           headers={'Accept': 'application/rdf+xml'}) as req:
            result = self.app_handle_request(req)
            with NamedTemporaryFile(delete=False) as fobj:
                try:
                    fobj.write(result)
                    fobj.close()
                    graph = default_graph()
                    graph.load('file://' + fobj.name, rdf_format='xml')
                finally:
                    os.unlink(fobj.name)

    def test_eac_download_filename(self):
        with self.admin_access.web_request() as req:
            cnx = req.cnx
            agent = testutils.authority_record(cnx, u'jim')
            for ark, expected_filename in (
                (u"", "EAC_{0}.xml".format(agent.eid)),
                (u"ZZZ/4242", "EAC_ZZZ-4242.xml".format(agent.eid)),
            ):
                agent.cw_set(ark=ark)
                view = self.vreg['views'].select('eac.export', req, agent.as_rset())
                view.set_request_content_type()
                self.assertEqual(
                    view._cw.headers_out.getRawHeaders('content-disposition'),
                    ['attachment;filename="{0}"'.format(expected_filename)],
                )

    def test_agent_place_as_concept_view(self):
        with self.admin_access.web_request() as req:
            cnx = req.cnx
            agent_place = cnx.create_entity('AgentPlace', name=u"""çàè€É'"><$""")
            content = agent_place.view('saem.agent_place_as_concept')
            self.assertEqual(content,
                             u"<strong><span>çàè€É&#39;&quot;&gt;&lt;$</span></strong>")

    def test_listitem(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWGroup', name=u'a')
            content = entity.view('saem.listitem', tabid='blah')
            self.assertIn(entity.view('oneline'), content)
            self.assertIn('__redirectparams=tab%3Dblah&amp', content)

    def test_prov_activities(self):
        with self.admin_access.web_request() as req:
            record = testutils.authority_record(req, u'My record')
            req.create_entity('Activity', agent=u'007', generated=record)
            req.cnx.commit()
            table_html = record.view('prov.activity-generated')
            self.assertIn('admin', table_html)
            self.assertIn('007', table_html)

    def test_related_entities_navigation(self):
        with self.admin_access.web_request() as req:
            ou = testutils.organization_unit(req, u'Archival Dptmt', ['archival'])
            testutils.authority_record(req.cnx, u'bob', reverse_use_authorityrecord=ou)
            testutils.authority_record(req.cnx, u'alice', reverse_use_authorityrecord=ou)
            req.cnx.commit()

            vid = 'saem.ou.authorityrecords'
            view = req.vreg['views'].select(vid, req, rset=ou.as_rset())
            html = view.render()
            nav_link = view.page_navigation_url(None, None, {})

            expected_divid = '{}{}'.format(vid.replace('.', '_'), ou.eid)
            self.assertIn('<div id="{}">'.format(expected_divid), html)
            self.assertIn('#' + expected_divid, nav_link)
            params = json.loads('{{ {} }}'.format(nav_link.split('{')[1].split('}')[0]))
            self.assertEqual(params,
                             {'divid': expected_divid,
                              'vid': vid,
                              'rql': u'Any X WHERE X eid {}'.format(ou.eid)})

    def test_email_address(self):
        with self.admin_access.cnx() as cnx:
            address = cnx.create_entity('EmailAddress', address=u'admin@cubicweb.org')
            self.assertEqual(address.dc_title(), u'admin@cubicweb.org')


class SEDAViewsTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def test_seda_get_related_version(self):
        """Check that we get correct results when asking for `draft`, `published`, `replaced`
        version of a profile."""
        with self.admin_access.web_request() as req:
            profile1 = testutils.setup_profile(req, title=u'Profile 1')
            req.cnx.commit()
            profile1.cw_adapt_to('IWorkflowable').fire_transition('publish')
            req.cnx.commit()
            profile1.cw_clear_all_caches()
            profile2 = testutils.setup_profile(req, title=u'Profile 2', new_version_of=profile1)
            req.cnx.commit()
            profile2.cw_adapt_to('IWorkflowable').fire_transition('publish')
            req.cnx.commit()
            profile2.cw_clear_all_caches()
            profile3 = testutils.setup_profile(req, title=u'Profile 3', new_version_of=profile2)
            req.cnx.commit()
            profile3.cw_adapt_to('IWorkflowable').fire_transition('publish')
            req.cnx.commit()
            profile3.cw_clear_all_caches()
            profile4 = testutils.setup_profile(req, title=u'Profile 4', new_version_of=profile3)
            req.cnx.commit()

            def unwrap_generator(gen):
                try:
                    return next(iter(gen))
                except StopIteration:
                    return None

            # Draft profile
            box4 = self.vreg['ctxcomponents'].select('saem.seda.relatedprofiles', req,
                                                     entity=profile4)
            self.assertEqual(unwrap_generator(box4.predecessor()).eid, profile3.eid)
            self.assertIsNone(unwrap_generator(box4.current_version(state=u'published')))
            self.assertIsNone(unwrap_generator(box4.current_version(state=u'draft')))
            # Published profile
            box3 = self.vreg['ctxcomponents'].select('saem.seda.relatedprofiles', req,
                                                     entity=profile3)
            self.assertEqual(unwrap_generator(box3.predecessor()).eid, profile2.eid)
            self.assertIsNone(unwrap_generator(box3.current_version(state=u'published')))
            self.assertEqual(unwrap_generator(box3.current_version(state=u'draft')).eid,
                             profile4.eid)
            # Deprecated profile
            box2 = self.vreg['ctxcomponents'].select('saem.seda.relatedprofiles', req,
                                                     entity=profile2)
            self.assertIsNone(unwrap_generator(box2.predecessor()))
            self.assertEqual(unwrap_generator(box2.current_version(state=u'published')).eid,
                             profile3.eid)
            self.assertEqual(unwrap_generator(box2.current_version(state=u'draft')).eid,
                             profile4.eid)
            # Older deprecated profile
            box1 = self.vreg['ctxcomponents'].select('saem.seda.relatedprofiles', req,
                                                     entity=profile1)
            self.assertIsNone(unwrap_generator(box1.predecessor()))
            self.assertEqual(unwrap_generator(box1.current_version(state=u'published')).eid,
                             profile3.eid)
            self.assertEqual(unwrap_generator(box1.current_version(state=u'draft')).eid,
                             profile4.eid)


class ArkViewsTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def _check_entity_from_path(self, cnx, path, etype, parent_ark=None,
                                **kwargs):
        """Make sure an entity with `etype` and `kwargs` attributes values
        exists and matches ARK identifier found in `path`.
        """
        entity = cnx.find(etype, **kwargs).one()
        self.assertTrue(path.startswith('ark:/'), path)
        self.assertEqual(entity.ark, path[len('ark:/'):])
        if parent_ark is not None:
            self.assertTrue(entity.ark.startswith(parent_ark + '/'),
                            entity.ark)

    def test_ark_authority_record_creation(self):
        with self.admin_access.cnx() as cnx:
            org = testutils.authority_with_naa(cnx)
            naa_eid = org.ark_naa[0].eid
            cnx.commit()
        with self.admin_access.web_request() as req:
            akind = req.cnx.find('AgentKind', name=u'person').one()
            record = self.vreg['etypes'].etype_class('AuthorityRecord')(req)
            record.eid = 'A'
            fields = {
                role_name('record_id', 'subject'): u'test',
                role_name('ark', 'subject'): u'',
                role_name('agent_kind', 'subject'): str(akind.eid),
                role_name('ark_naa', 'subject'): str(naa_eid),
            }
            name = self.vreg['etypes'].etype_class('NameEntry')(req)
            name.eid = 'B'
            name_fields = {
                role_name('parts', 'subject'): u'007',
                role_name('name_entry_for', 'subject'): u'A',
            }
            req.form = self.fake_form('edition', entity_field_dicts=[(record, fields),
                                                                     (name, name_fields)])
            path = self.expect_redirect_handle_request(req)[0]
            self._check_entity_from_path(req, path, 'AuthorityRecord',
                                         record_id=u'test')

    def test_ark_agent_creation(self):
        with self.admin_access.web_request() as req:
            org = testutils.authority_with_naa(req)
            agent = self.vreg['etypes'].etype_class('Agent')(req)
            agent.eid = 'A'
            fields = {role_name('name', 'subject'): u'007',
                      role_name('ark', 'subject'): u'',
                      role_name('authority', 'subject'): str(org.eid)}
            req.form = self.fake_form('edition', entity_field_dicts=[(agent, fields)])
            path = self.expect_redirect_handle_request(req)[0]
            self._check_entity_from_path(req, path, 'Agent',
                                         parent_ark=org.ark,
                                         name=u'007')

    def test_ark_scheme_creation(self):
        with self.admin_access.web_request() as req:
            scheme = self.vreg['etypes'].etype_class('ConceptScheme')(req)
            scheme.eid = 'A'
            fields = {role_name('title', 'subject'): u'test',
                      role_name('ark', 'subject'): u'',
                      role_name('ark_naa', 'subject'): text_type(testutils.naa(req.cnx).eid)}
            req.form = self.fake_form('edition', entity_field_dicts=[(scheme, fields)])
            path = self.expect_redirect_handle_request(req)[0]
            self._check_entity_from_path(req, path, 'ConceptScheme',
                                         title=u'test')

    def test_ark_concept_creation_form(self):
        # test addition of a concept by specifying in_scheme in form
        with self.admin_access.web_request() as req:
            scheme = req.cnx.create_entity('ConceptScheme', ark_naa=testutils.naa(req.cnx))
            concept = self.vreg['etypes'].etype_class('Concept')(req)
            concept.eid = 'A'
            concept_fields = {role_name('in_scheme', 'subject'): str(scheme.eid),
                              role_name('definition', 'subject'): u'thing',
                              role_name('ark', 'subject'): u''}
            label = self.vreg['etypes'].etype_class('Label')(req)
            label.eid = 'B'
            label_fields = {role_name('label', 'subject'): u'Hello',
                            role_name('label_of', 'subject'): 'A'}
            req.form = self.fake_form('edition', entity_field_dicts=[(concept, concept_fields),
                                                                     (label, label_fields)])
            path = self.expect_redirect_handle_request(req)[0]
            self._check_entity_from_path(req, path, 'Concept',
                                         parent_ark=scheme.ark,
                                         definition=u'thing')

    def test_ark_url_rewrite(self):
        with self.admin_access.web_request() as req:
            rewriter = self.vreg['urlrewriting'].select('schemabased', req)
            _pmid, rset = rewriter.rewrite(req, u'/ark:/JoE/Dalton')
            self.assertEqual(rset.printable_rql(), 'Any X WHERE X ark "JoE/Dalton"')


class AssignArkWebServiceTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def test_not_authenticated(self):
        with self.new_access('anon').web_request(headers={'Accept': 'application/json'},
                                                 method='POST') as req:
            result = self.app_handle_request(req, 'ark')
            self.assertEqual(json.loads(result),
                             [{'error': 'This service requires authentication.'}])

    def test_authenticated_missing_organization(self):
        with self.admin_access.web_request(
                'ark',
                headers={'Accept': 'application/json'}, method='POST') as req:
            result = self.app_handle_request(req)
            self.assertEqual(json.loads(result),
                             [{'error': 'Missing required "organization" query parameter.'}])

    def test_authenticated_organization_method_not_post(self):
        with self.admin_access.cnx() as cnx:
            org_ark = testutils.authority_with_naa(cnx).ark
        params = {'organization': org_ark}
        with self.admin_access.web_request(
                headers={'Accept': 'application/json'}, **params) as req:
            result = self.app_handle_request(req, 'ark')
            self.assertEqual(json.loads(result),
                             [{'error': 'This service is only accessible using POST.'}])

    def test_authenticated_organization_does_not_exist_method_post(self):
        expected_msg = 'No organization matching identifier "{0}".'
        # Organization does not exist.
        bad_params = {'organization': u'0'}
        with self.admin_access.web_request(
                headers={'Accept': 'application/json'},
                method='POST', **bad_params) as req:
            result = self.app_handle_request(req, 'ark')
            self.assertEqual(json.loads(result),
                             [{'error': expected_msg.format(0)}])
        # Specified ark does not match an Organization entity.
        with self.admin_access.cnx() as cnx:
            agent_ark = testutils.agent(cnx, u'bob').ark
        bad_params = {'organization': agent_ark}
        with self.admin_access.web_request(
                headers={'Accept': 'application/json'},
                method='POST', **bad_params) as req:
            result = self.app_handle_request(req, 'ark')
            self.assertEqual(json.loads(result),
                             [{'error': expected_msg.format(agent_ark)}])

    def test_authenticated_organization_has_no_naa_method_post(self):
        with self.admin_access.cnx() as cnx:
            org_ark = cnx.create_entity(
                'Organization', name=u'dummy', ark=u'123/dummy').ark
            cnx.commit()
        bad_params = {'organization': org_ark}
        with self.admin_access.web_request(
                headers={'Accept': 'application/json'},
                method='POST', **bad_params) as req:
            result = self.app_handle_request(req, 'ark')
            expected_msg = 'Organization "{0}" cannot assign ARK identifiers.'
            self.assertEqual(json.loads(result),
                             [{'error': expected_msg.format(org_ark)}])

    def test_ok(self):
        with self.admin_access.cnx() as cnx:
            org_ark = testutils.authority_with_naa(cnx).ark
            cnx.commit()

        def check_ark(response_body):
            json_response = json.loads(result)
            self.assertEqual(len(json_response), 1)
            self.assertEqual(list(json_response[0].keys()), ['ark'])
            ark = json_response[0]['ark']
            self.assertTrue(testutils.ARK_RGX.match(ark), ark)

        params = {'organization': org_ark}
        with self.admin_access.web_request(
                headers={'Accept': 'application/json'},
                method='POST', **params) as req:
            result = self.app_handle_request(req, 'ark')
            check_ark(result)
        params = {'organization': 'ark:/' + org_ark}
        with self.admin_access.web_request(
                headers={'Accept': 'application/json'},
                method='POST', **params) as req:
            result = self.app_handle_request(req, 'ark')
            check_ark(result)


class ListArchivalUnitsServiceTC(CubicWebTC):

    configcls = PostgresApptestConfiguration
    maxDiff = None

    def test_ok(self):
        with self.admin_access.cnx() as cnx:
            org = testutils.authority_with_naa(cnx)
            archival_unit = testutils.organization_unit(
                cnx, u'Archival Dptmt', ['archival'],
                reverse_archival_unit=org,
                authority=org)
            dep_unit = testutils.organization_unit(
                cnx, u'Deposit Dptmt', ['deposit'],
                authority=org)
            deprecated_dep_unit = testutils.organization_unit(
                cnx, u'Deprecated Deposit Dptmt', ['deposit'],
                authority=org)
            testutils.organization_unit(
                cnx, u'Production Dptmt', ['producer'],
                authority=org)
            org2 = testutils.authority_with_naa(cnx, name=u'Another authority',
                                                archival_unit=archival_unit)
            dep2_unit = testutils.organization_unit(
                cnx, u'Org 2 Deposit Dptmt', ['deposit'],
                authority=org2)
            cnx.commit()
            deprecated_dep_unit.cw_adapt_to('IWorkflowable').fire_transition('deprecate')
            cnx.commit()
            # preload attributes accessed in later connection
            org.ark, org2.ark
            archival_unit.ark, dep_unit.ark, deprecated_dep_unit.ark, dep2_unit.ark

        with self.admin_access.web_request(
                headers={'Accept': 'application/json'}) as req:
            result = json.loads(self.app_handle_request(req, 'archival-units'))
            expected = [
                {
                    'ark': archival_unit.ark,
                    'name': 'Archival Dptmt',
                    'state': 'published',
                    'url': archival_unit.absolute_url(),
                    'authority': {
                        'ark': org.ark,
                        'name': org.name,
                        'url': org.absolute_url(),
                    },
                    'deposit-units': [
                        {
                            'ark': dep_unit.ark,
                            'name': dep_unit.name,
                            'state': 'published',
                            'url': dep_unit.absolute_url(),
                            'authority': {
                                'ark': org.ark,
                                'name': org.name,
                                'url': org.absolute_url(),
                            },
                        },
                        {
                            'ark': deprecated_dep_unit.ark,
                            'name': deprecated_dep_unit.name,
                            'state': 'deprecated',
                            'url': deprecated_dep_unit.absolute_url(),
                            'authority': {
                                'ark': org.ark,
                                'name': org.name,
                                'url': org.absolute_url(),
                            },
                        },
                        {
                            'ark': dep2_unit.ark,
                            'name': dep2_unit.name,
                            'state': 'published',
                            'url': dep2_unit.absolute_url(),
                            'authority': {
                                'ark': org2.ark,
                                'name': org2.name,
                                'url': org2.absolute_url(),
                            },
                        },
                    ],
                },
            ]
            result[0]['deposit-units'].sort()
            expected[0]['deposit-units'].sort()
            self.assertEqual(result, expected)


class TimelineViewsTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def test_timeline_json(self):
        with self.admin_access.cnx() as cnx:
            ar1 = testutils.authority_record(
                cnx, u'1',
                start_date=date(1987, 6, 1), end_date=date(1988, 4, 2))
            ar2 = testutils.authority_record(
                cnx, u'2',
                start_date=date(1989, 1, 1), end_date=date(1990, 4, 3))
            ar3 = testutils.authority_record(
                cnx, u'2',
                start_date=date(1991, 1, 1), end_date=date(1992, 4, 3))
            cnx.create_entity('ChronologicalRelation',
                              description=u'avant avant',
                              chronological_predecessor=ar1,
                              chronological_successor=ar3)
            cnx.create_entity('ChronologicalRelation',
                              description=u'juste avant',
                              chronological_predecessor=ar2,
                              chronological_successor=ar3)
            cnx.commit()
        with self.admin_access.web_request() as req:
            ar3 = req.entity_from_eid(ar3.eid)
            ar2 = req.entity_from_eid(ar2.eid)
            ar1 = req.entity_from_eid(ar1.eid)
            data = json.loads(ar3.view(
                'saem.authorityrecord-timeline-json', w=None))
            predecessors = [x for x in data['timeline']['date']
                            if x['tag'] == u'chronological_predecessor']
            self.assertEqual(len(predecessors), 2)
            self.assertCountEqual([p['headline'] for p in predecessors],
                                  [ar2.view('incontext', w=None),
                                   ar1.view('incontext', w=None)])


class BaseViewsTC(CubicWebTC):
    """Test for non-business views."""
    configcls = PostgresApptestConfiguration

    def expected_notfound_html(self, etype, addlink=False, kwargs=None):
        parts = [
            '<div class="searchMessage">No entity of type &quot;{}&quot; yet.'.format(etype)
        ]
        if addlink:
            path = 'add/' + etype
            if kwargs:
                path = '{}?{}'.format(
                    path, '&'.join(['{}={}'.format(k, v) for k, v in kwargs.items()]))
            parts.append(
                ' <a href="http://testing.fr/cubicweb/{}">Create one?</a>'.format(path)
            )
        parts.append('</div>')
        return ''.join(parts)

    def test_noresult_sedalib(self):
        with self.admin_access.web_request() as req:
            result = self.app_handle_request(req, path='/sedalib')
            kwargs = {'unit_type': 'unit_content'}
            self.assertIn(self.expected_notfound_html('SEDAArchiveUnit', True, kwargs), result)
        with self.new_access('anon').web_request() as req:
            result = self.app_handle_request(req, path='/sedalib')
            self.assertIn(self.expected_notfound_html('SEDAArchiveUnit'), result)

    def test_no_notfound_when_no_entity(self):
        """Check that we do not get 404 at /<etype> when no entity of <etype>
        exists yet.
        """
        with self.admin_access.web_request() as req:
            # Request /bookmark, usually there's none on a fresh instance.
            result = self.app_handle_request(req, path='/bookmark')
            self.assertIn(self.expected_notfound_html('Bookmark', True), result)
        with self.new_access('anon').web_request() as req:
            result = self.app_handle_request(req, path='/bookmark')
            self.assertIn(self.expected_notfound_html('Bookmark'), result)

    def test_true_404(self):
        """Check 404 is returned on unknown entity type."""
        with self.admin_access.web_request() as req:
            self.app_handle_request(req, path='/zorglub')
            self.assertEqual(req.status_out, 404)


class SortableListViewMixinTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    class FakeListView(object):

        @staticmethod
        def related_rset(entity):
            return entity.related('mandate_agent', role='object')

    class SortableFakeListView(SortableListViewMixin, FakeListView):
        pass

    def test_related_rset_sorted_by_index(self):
        with self.admin_access.cnx() as cnx:
            arecord = testutils.authority_record(cnx, u'record')

            m1 = cnx.create_entity('Mandate', term=u'1', mandate_agent=arecord)
            m2 = cnx.create_entity('Mandate', term=u'2', mandate_agent=arecord)
            m3 = cnx.create_entity('Mandate', term=u'3', mandate_agent=arecord)
            cnx.commit()
            fake_view = self.SortableFakeListView()
            expected = [m1.eid, m2.eid, m3.eid]
            actual = [e.eid for e in fake_view.related_rset(arecord).entities()]
            self.assertEqual(expected, actual)

            m1.cw_set(index=2)
            m2.cw_set(index=3)
            m3.cw_set(index=1)
            cnx.commit()
            expected = [m3.eid, m1.eid, m2.eid]
            actual = [e.eid for e in fake_view.related_rset(arecord).entities()]
            self.assertEqual(expected, actual)


class SortableAjaxControllerTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            arecord = testutils.authority_record(cnx, u'record')

            self.eid0 = cnx.create_entity('Mandate', term=u'1', mandate_agent=arecord).eid
            self.eid1 = cnx.create_entity('Mandate', term=u'2', mandate_agent=arecord).eid
            self.eid2 = cnx.create_entity('Mandate', term=u'3', mandate_agent=arecord).eid
            self.eids = [self.eid0, self.eid1, self.eid2]
            cnx.commit()

    def test_sort_relation_targets_sorts_sortable(self):
        sorted_eids = [str(eid) for eid in (self.eid1, self.eid0, self.eid2)]
        with self.remote_calling('sort_relation_targets', 'mandate_agent', sorted_eids) as (_, req):
            indexes = [req.entity_from_eid(eid).index for eid in self.eids]
            self.assertEqual([1, 0, 2], indexes)
        sorted_eids = [str(eid) for eid in (self.eid2, self.eid0, self.eid1)]
        with self.remote_calling('sort_relation_targets', 'mandate_agent', sorted_eids) as (_, req):
            indexes = [req.entity_from_eid(eid).index for eid in self.eids]
            self.assertEqual([1, 2, 0], indexes)


if __name__ == '__main__':
    unittest.main()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref unit tests for entities.container"""

from cubicweb.devtools import PostgresApptestConfiguration
from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_saem_ref.entities import container

import testutils


def setUpModule():
    testutils.startpgcluster(__file__)


def sort_container(container_def):
    for k, v in container_def:
        yield k, sorted(v)


class ContainerTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def test_authorityrecord_container(self):
        # line below should be copied from entities.container.registration_callback
        container_def = container.authority_record_container_def(self.schema)
        container_def = dict(sort_container(container_def))
        self.assertEqual(container_def,
                         {'AgentFunction': [('function_agent', 'subject')],
                          'AgentPlace': [('place_agent', 'subject')],
                          'Citation': [('has_citation', 'object')],
                          'EACOtherRecordId': [('eac_other_record_id_of', 'subject')],
                          'EACResourceRelation': [('resource_relation_agent', 'subject')],
                          'EACSource': [('source_agent', 'subject')],
                          'GeneralContext': [('general_context_of', 'subject')],
                          'History': [('history_agent', 'subject')],
                          'LegalStatus': [('legal_status_agent', 'subject')],
                          'Mandate': [('mandate_agent', 'subject')],
                          'NameEntry': [('name_entry_for', 'subject')],
                          'Occupation': [('occupation_agent', 'subject')],
                          'PostalAddress': [('place_address', 'object')],
                          'Structure': [('structure_agent', 'subject')]})
        entity = self.vreg['etypes'].etype_class('AuthorityRecord')(self)
        self.assertIsNotNone(entity.cw_adapt_to('IContainer'))
        self.assertIsNone(entity.cw_adapt_to('IContained'))

    def test_scheme_container(self):
        # line below should be copied from entities.container.registration_callback
        container_def = container.scheme_container_def(self.schema)
        container_def = dict(sort_container(container_def))
        self.assertEqual(container_def,
                         {'Concept': [('in_scheme', 'subject')]})
        entity = self.vreg['etypes'].etype_class('ConceptScheme')(self)
        self.assertIsNotNone(entity.cw_adapt_to('IContainer'))
        self.assertIsNone(entity.cw_adapt_to('IContained'))

    def test_concept_container(self):
        # line below should be copied from entities.container.registration_callback
        container_def = container.concept_container_def(self.schema)
        container_def = dict(sort_container(container_def))
        self.assertEqual(container_def,
                         {'Label': [('label_of', 'subject')]})
        entity = self.vreg['etypes'].etype_class('Concept')(self)
        self.assertIsNotNone(entity.cw_adapt_to('IContainer'))
        # Concept is both container and contained :
        self.assertIsNotNone(entity.cw_adapt_to('IContained'))


class TreeTC(CubicWebTC):

    configcls = PostgresApptestConfiguration

    def test_seda_profile_clone(self):
        """Functional test for SEDA profile cloning."""
        with self.admin_access.repo_cnx() as cnx:
            transfer = testutils.setup_profile(cnx)
            unit, unit_alt, unit_alt_seq = testutils.create_archive_unit(transfer)
            # don't add link from data object to transfer intentionally to force
            # going through the archive unit to clone it
            bdo = testutils.create_data_object(unit_alt_seq)
            cnx.commit()
            transfer.cw_adapt_to('IWorkflowable').fire_transition('publish')
            testutils.organization_unit(cnx, u'bob', archival_roles=['deposit'],
                                        use_profile=transfer)
            cnx.commit()

            clone = testutils.setup_profile(cnx, title=u'Clone', new_version_of=transfer)
            cnx.commit()

            # ark and cwuri should not have been copied
            self.assertNotEqual(clone.ark, transfer.ark)
            self.assertNotEqual(clone.cwuri, transfer.cwuri)

            # Everything else beside some explicitly skipped relations should
            # have been copied (but other parts of the test live in the seda
            # cube)
            self.assertEqual(clone.ark_naa[0].eid, transfer.ark_naa[0].eid)
            self.assertEqual(len(clone.reverse_use_profile), 0)

            # Ensure activites are not copied
            self.assertEqual(len(clone.reverse_generated), 1)
            self.assertEqual(len(clone.reverse_used), 1)
            self.assertNotEqual(clone.reverse_used[0].start, transfer.reverse_used[-1].start)

            # Ensure data object is cloned through data_object_reference_id and
            # container relation is properly handled
            bdo_clone = cnx.execute('Any MAX(X) WHERE X is SEDABinaryDataObject').one()
            self.assertNotEqual(bdo_clone.eid, bdo.eid)
            bdo.cw_clear_all_caches()
            self.assertEqual([e.eid for e in bdo.container], [transfer.eid])
            self.assertEqual([e.eid for e in bdo_clone.container], [clone.eid])

            self.assertFalse(cnx.execute('Any X GROUPBY X WHERE X container C HAVING COUNT(C) > 1'))
            self.assertFalse(cnx.execute('Any X WHERE NOT X container C'))


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref "compound"-related functionalities."""

from functools import partial

from cubicweb_compound import structure_def
from cubicweb_compound.entities import IContainer, IContained

# don't put Activity in a compound graph, it causes problem since it belongs to both
# SEDAArchiveTransfer tree and to concept/authority record tree, which don't have the same
# IContained implementation.
structure_def = partial(structure_def, skipetypes=('Activity',))


def authority_record_container_def(schema):
    """Define container for AuthorityRecord"""
    return structure_def(schema, 'AuthorityRecord').items()


def org_container_def(schema):
    return structure_def(schema, 'Organization').items()


def scheme_container_def(schema):
    """Define container for ConceptScheme"""
    skiprtypes = ('broader_concept', 'label_of')
    return structure_def(schema, 'ConceptScheme', skiprtypes=skiprtypes).items()


def concept_container_def(schema):
    """Define container for Concept"""
    return structure_def(schema, 'Concept').items()


def registration_callback(vreg):
    vreg.register(IContainer.build_class('Organization'))
    for etype, parent_relations in org_container_def(vreg.schema):
        IContained.register_class(vreg, etype, parent_relations)
    vreg.register(IContainer.build_class('AuthorityRecord'))
    for etype, parent_relations in authority_record_container_def(vreg.schema):
        IContained.register_class(vreg, etype, parent_relations)
    vreg.register(IContainer.build_class('ConceptScheme'))
    for etype, parent_relations in scheme_container_def(vreg.schema):
        if etype == 'Concept':
            # XXX turn parent_relations to a list to ensure broader_concept is considered first
            parent_relations = list(parent_relations)
            parent_relations.insert(0, ('broader_concept', 'subject'))
        IContained.register_class(vreg, etype, parent_relations)
    vreg.register(IContainer.build_class('Concept'))
    for etype, parent_relations in concept_container_def(vreg.schema):
        IContained.register_class(vreg, etype, parent_relations)



# coding: utf-8
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref site customizations"""

import pytz

from logilab.common.configuration import REQUIRED
from logilab.common.date import ustrftime
from logilab.common.decorators import monkeypatch

from cubicweb import (
    cwvreg,
    _,
)
from cubicweb.cwconfig import register_persistent_options
from cubicweb.uilib import PRINTERS
from cubicweb.entity import Entity
from cubicweb.web import request

from cubicweb_skos import rdfio
from cubicweb_skos.ccplugin import ImportSkosData
import cubicweb_seda as seda
import cubicweb_seda.dataimport as seda_dataimport

from . import (
    ark_naa,
    permanent_url,
    _massive_store_factory,
    _nohook_store_factory,
)

# this import is needed to take account of pg_trgm monkeypatches
# while executing cubicweb-ctl commands (db-rebuild-fti)
from . import pg_trgm  # noqa pylint: disable=unused-import


options = (
    ('default-ark-naa-who',
     {'type': 'string',
      'default': REQUIRED,
      'help': ('Name of default ARK Name Assigning Authority (who)'
               ' (only used at database initialization)'),
      'group': 'saem_ref',
      'level': 0}),
    ('default-ark-naa-what',
     {'type': 'int',
      'default': REQUIRED,
      'help': ('Number of default ARK Name Assigning Authority (what)'
               ' (only used at database initialization)'),
      'group': 'saem_ref',
      'level': 0}),
)


# override Entity.rest_path to use ark for entities which have one

_orig_rest_path = Entity.rest_path


@monkeypatch(Entity)
def rest_path(self, *args, **kwargs):
    """Return ark:/<ARK> is entity has an ark attribute."""
    if getattr(self, 'ark', None) is None:
        return _orig_rest_path(self, *args, **kwargs)
    return u'ark:/' + self.ark


# configure RDF generator to use ark based uri as canonical uris, and deactivating implicit
# 'same_as_urls' in this case

orig_same_as_uris = rdfio.RDFGraphGenerator.same_as_uris


@monkeypatch(rdfio.RDFGraphGenerator, methodname='same_as_uris')
@staticmethod
def same_as_uris(entity):
    if entity.cwuri.startswith('ark:'):
        return ()
    return orig_same_as_uris(entity)


@monkeypatch(rdfio.RDFGraphGenerator, methodname='canonical_uri')
@staticmethod
def canonical_uri(entity):
    return permanent_url(entity)


# deactivate date-format and datetime-format cw properties. This is because we do some advanced date
# manipulation such as allowing partial date and this is not generic enough to allow arbitrary
# setting of date and time formats

base_user_property_keys = cwvreg.CWRegistryStore.user_property_keys


@monkeypatch(cwvreg.CWRegistryStore)
def user_property_keys(self, withsitewide=False):
    props = base_user_property_keys(self, withsitewide)
    return [prop for prop in props if prop not in ('ui.date-format', 'ui.datetime-format')]


# customize display of TZDatetime

register_persistent_options((
    ('timezone',
     {'type': 'choice',
      'choices': pytz.common_timezones,
      'default': 'Europe/Paris',
      'help': _('timezone in which time should be displayed'),
      'group': 'ui', 'sitewide': True,
      }),
))


def print_tzdatetime_local(value, req, *args, **kwargs):
    tz = pytz.timezone(req.property_value('ui.timezone'))
    value = value.replace(tzinfo=pytz.utc).astimezone(tz)
    return ustrftime(value, req.property_value('ui.datetime-format'))


PRINTERS['TZDatetime'] = print_tzdatetime_local


# configure c-c skos-import command's factories to use with proper metadata generator ##############

ImportSkosData.cw_store_factories['massive'] = _massive_store_factory
ImportSkosData.cw_store_factories['nohook'] = _nohook_store_factory


# override seda's scheme initialization to set ark on each scheme, and to use an ark enabled store

@monkeypatch(seda_dataimport)
def init_seda_scheme(cnx, title, _count=[]):
    description = u'edition 2009' if title.startswith('SEDA :') else None
    who = cnx.vreg.config['default-ark-naa-who']
    what = cnx.vreg.config['default-ark-naa-what']
    naa = ark_naa(cnx, who, what)
    return cnx.create_entity('ConceptScheme', title=title, description=description,
                             ark_naa=naa)


@monkeypatch(seda_dataimport)
def get_store(cnx):
    from .sobjects import SAEMMetadataGenerator
    what = cnx.vreg.config['default-ark-naa-what']
    metagen = SAEMMetadataGenerator(cnx, naa_what=what)
    if cnx.repo.system_source.dbdriver == 'postgres':
        from cubicweb.dataimport.massive_store import MassiveObjectStore
        return MassiveObjectStore(cnx, metagen=metagen, eids_seq_range=1000)
    else:
        from cubicweb.dataimport.stores import NoHookRQLObjectStore
        return NoHookRQLObjectStore(cnx, metagen=metagen)


# configure seda compound graph to discard Activity and its relations, else it
# causes problem because it belongs to several graphs with different compound
# implementation (using "container" relation or not)
seda.GRAPH_SKIP_ETYPES.add('Activity')
# also, the new_version_of relation should not be considered as part of the
# graph (as for e.g. container or clone_of)
for rtype in ('new_version_of', 'use_profile'):
    seda.GRAPH_SKIP_RTYPES.add(rtype)
    Entity.cw_skip_copy_for.append((rtype, 'subject'))
    Entity.cw_skip_copy_for.append((rtype, 'object'))


@monkeypatch(request._CubicWebRequestBase)
def negotiated_language(self):
    # Force language to fr since in http-negociation mode there is no way to
    # force french language with a browser configured in english.
    # This is currently the easiest way to force the language of an instance.
    return 'fr'
